@@ -22,13 +22,16 @@ import gzip
 import io
 import math
 import os
+import shlex
 import shutil
+import stat
 import tempfile
 
 from ironic_lib import disk_utils
 from ironic_lib import exception
 from ironic_lib import utils
 from oslo_concurrency import processutils
+from oslo_config import cfg
 from oslo_log import log
 from oslo_utils import excutils
 from oslo_utils import units
@@ -36,10 +39,12 @@ from oslo_utils import uuidutils
 import requests
 
 from ironic_python_agent import errors
-from ironic_python_agent.extensions import image
+from ironic_python_agent import hardware
+from ironic_python_agent import utils as ipa_utils
 
 
 LOG = log.getLogger()
+CONF = cfg.CONF
 
 MAX_CONFIG_DRIVE_SIZE_MB = 64
 
@@ -63,13 +68,27 @@ def get_configdrive(configdrive, node_uuid, tempdir=None):
     # Check if the configdrive option is a HTTP URL or the content directly
     is_url = utils.is_http_url(configdrive)
     if is_url:
+        verify, cert = ipa_utils.get_ssl_client_options(CONF)
+        timeout = CONF.image_download_connection_timeout
+        # TODO(dtantsur): support proxy parameters from instance_info
         try:
-            data = requests.get(configdrive).content
+            resp = requests.get(configdrive, verify=verify, cert=cert,
+                                timeout=timeout)
         except requests.exceptions.RequestException as e:
             raise exception.InstanceDeployFailure(
                 "Can't download the configdrive content for node %(node)s "
                 "from '%(url)s'. Reason: %(reason)s" %
                 {'node': node_uuid, 'url': configdrive, 'reason': e})
+
+        if resp.status_code >= 400:
+            raise exception.InstanceDeployFailure(
+                "Can't download the configdrive content for node %(node)s "
+                "from '%(url)s'. Got status code %(code)s, response "
+                "body %(body)s" %
+                {'node': node_uuid, 'url': configdrive,
+                 'code': resp.status_code, 'body': resp.text})
+
+        data = resp.content
     else:
         data = configdrive
 
@@ -423,7 +442,7 @@ def create_config_drive_partition(node_uuid, device, configdrive):
                     device, new_part.pop())
             else:
                 try:
-                    config_drive_part = image._get_partition(device, part_uuid)
+                    config_drive_part = get_partition(device, part_uuid)
                 except errors.DeviceNotFound:
                     msg = ('Failed to create config drive on disk %(disk)s '
                            'for node %(node)s. Partition with UUID %(uuid)s '
@@ -489,3 +508,92 @@ def _is_disk_larger_than_max_size(device, node_uuid):
     disksize_mb = int(disksize_bytes.strip()) // 1024 // 1024
 
     return disksize_mb > MAX_DISK_SIZE_MB_SUPPORTED_BY_MBR
+
+
+def get_partition(device, uuid):
+    """Find the partition of a given device."""
+    LOG.debug("Find the partition %(uuid)s on device %(dev)s",
+              {'dev': device, 'uuid': uuid})
+
+    try:
+        ipa_utils.rescan_device(device)
+        lsblk = utils.execute(
+            'lsblk', '-PbioKNAME,UUID,PARTUUID,TYPE,LABEL', device)
+        report = lsblk[0]
+        for line in report.split('\n'):
+            part = {}
+            # Split into KEY=VAL pairs
+            vals = shlex.split(line)
+            for key, val in (v.split('=', 1) for v in vals):
+                part[key] = val.strip()
+            # Ignore non partition
+            if part.get('TYPE') not in ['md', 'part']:
+                # NOTE(TheJulia): This technically creates an edge failure
+                # case where a filesystem on a whole block device sans
+                # partitioning would behave differently.
+                continue
+
+            if part.get('UUID') == uuid:
+                LOG.debug("Partition %(uuid)s found on device "
+                          "%(dev)s", {'uuid': uuid, 'dev': device})
+                return '/dev/' + part.get('KNAME')
+            if part.get('PARTUUID') == uuid:
+                LOG.debug("Partition %(uuid)s found on device "
+                          "%(dev)s", {'uuid': uuid, 'dev': device})
+                return '/dev/' + part.get('KNAME')
+            if part.get('LABEL') == uuid:
+                LOG.debug("Partition %(uuid)s found on device "
+                          "%(dev)s", {'uuid': uuid, 'dev': device})
+                return '/dev/' + part.get('KNAME')
+        else:
+            # NOTE(TheJulia): We may want to consider moving towards using
+            # findfs in the future, if we're comfortable with the execution
+            # and interaction. There is value in either way though.
+            # NOTE(rg): alternative: blkid -l -t UUID=/PARTUUID=
+            try:
+                findfs, stderr = utils.execute('findfs', 'UUID=%s' % uuid)
+                return findfs.strip()
+            except processutils.ProcessExecutionError as e:
+                LOG.debug('First fallback detection attempt for locating '
+                          'partition via UUID %(uuid)s failed. '
+                          'Error: %(err)s',
+                          {'uuid': uuid,
+                           'err': e})
+                try:
+                    findfs, stderr = utils.execute(
+                        'findfs', 'PARTUUID=%s' % uuid)
+                    return findfs.strip()
+                except processutils.ProcessExecutionError as e:
+                    LOG.debug('Secondary fallback detection attempt for '
+                              'locating partition via UUID %(uuid)s failed. '
+                              'Error: %(err)s',
+                              {'uuid': uuid,
+                               'err': e})
+
+            # Last fallback: In case we cannot find the partition by UUID
+            # and the deploy device is an md device, we check if the md
+            # device has a partition (which we assume to contain the root fs).
+            if hardware.is_md_device(device):
+                md_partition = device + 'p1'
+                if (os.path.exists(md_partition)
+                        and stat.S_ISBLK(os.stat(md_partition).st_mode)):
+                    LOG.debug("Found md device with partition %s",
+                              md_partition)
+                    return md_partition
+                else:
+                    LOG.debug('Could not find partition %(part)s on md '
+                              'device %(dev)s',
+                              {'part': md_partition,
+                               'dev': device})
+
+            # Partition not found, time to escalate.
+            error_msg = ("No partition with UUID %(uuid)s found on "
+                         "device %(dev)s" % {'uuid': uuid, 'dev': device})
+            LOG.error(error_msg)
+            raise errors.DeviceNotFound(error_msg)
+    except processutils.ProcessExecutionError as e:
+        error_msg = ('Finding the partition with UUID %(uuid)s on '
+                     'device %(dev)s failed with %(err)s' %
+                     {'uuid': uuid, 'dev': device, 'err': e})
+        LOG.error(error_msg)
+        raise errors.CommandExecutionError(error_msg)
