@@ -26,7 +26,6 @@ import sys
 import tarfile
 import time
 
-from ironic_lib import disk_utils
 from ironic_lib import utils as ironic_utils
 from oslo_concurrency import processutils
 from oslo_config import cfg
@@ -79,8 +78,6 @@ COLLECT_LOGS_COMMANDS = {
 
 DEVICE_EXTRACTOR = re.compile(r'^(?:(.*\d)p|(.*\D))(?:\d+)$')
 
-PARTED_TABLE_TYPE_REGEX = re.compile(r'^.*partition\s+table\s*:\s*(gpt|msdos)',
-                                     re.IGNORECASE)
 
 _EARLY_LOG_BUFFER = []
 
@@ -554,11 +551,19 @@ def collect_system_logs(journald_max_lines=None):
 
     io_dict = {}
     file_list = []
+    log_locations = [CONF.log_file, CONF.log_dir]
     if is_journalctl_present():
         io_dict['journal'] = get_journalctl_output(lines=journald_max_lines)
+        for log_loc in log_locations:
+            if log_loc and os.path.exists(log_loc):
+                file_list.append(log_loc)
     else:
         try_get_command_output(io_dict, 'dmesg', ['dmesg'])
         file_list.append('/var/log')
+        for log_loc in log_locations:
+            if (log_loc and os.path.exists(log_loc)
+                    and not log_loc.startswith('/var/log')):
+                file_list.append(log_loc)
 
     for name, cmd in COLLECT_LOGS_COMMANDS.items():
         try_get_command_output(io_dict, name, cmd)
@@ -729,44 +734,6 @@ def get_partition_table_type_from_specs(node):
     return 'gpt' if disk_label == 'gpt' else 'msdos'
 
 
-def scan_partition_table_type(device):
-    """Get partition table type, msdos or gpt.
-
-    :param device_name: the name of the device
-    :return: msdos, gpt or unknown
-    """
-    out, _u = execute('parted', '-s', device, '--', 'print')
-    out = out.splitlines()
-
-    for line in out:
-        m = PARTED_TABLE_TYPE_REGEX.match(line)
-        if m:
-            return m.group(1)
-
-    LOG.warning("Unable to get partition table type for device %s.",
-                device)
-
-    return 'unknown'
-
-
-def get_efi_part_on_device(device):
-    """Looks for the efi partition on a given device.
-
-    A boot partition on a GPT disk is assumed to be an EFI partition as well.
-
-    :param device: lock device upon which to check for the efi partition
-    :return: the efi partition or None
-    """
-    is_gpt = scan_partition_table_type(device) == 'gpt'
-    for part in disk_utils.list_partitions(device):
-        flags = {x.strip() for x in part['flags'].split(',')}
-        if 'esp' in flags or ('boot' in flags and is_gpt):
-            LOG.debug("Found EFI partition %s on device %s.", part, device)
-            return part['number']
-    else:
-        LOG.debug("No efi partition found on device %s", device)
-
-
 _LARGE_KEYS = frozenset(['configdrive', 'system_logs'])
 
 
@@ -869,8 +836,7 @@ def create_partition_table(dev_name, partition_table_type):
     :raises: CommandExecutionError if an error is encountered while
              attempting to create the partition table.
     """
-    LOG.info("Creating partition table on {}".format(
-        dev_name))
+    LOG.info("Creating partition table on %s", dev_name)
     try:
         execute('parted', dev_name, '-s', '--',
                 'mklabel', partition_table_type)
@@ -917,3 +883,21 @@ class StreamingClient:
         except requests.RequestException as exc:
             raise errors.CommandExecutionError(
                 "Unable to read data from %s: %s" % (url, exc))
+
+
+def rescan_device(device):
+    """Force the device to be rescanned
+
+    :param device: device upon which to rescan and update
+                   kernel partition records.
+    """
+    try:
+        execute('partx', '-a', device, attempts=3, delay_on_retry=True)
+    except processutils.ProcessExecutionError:
+        LOG.warning("Couldn't re-read the partition table "
+                    "on device %s", device)
+    try:
+        execute('udevadm', 'settle')
+    except processutils.ProcessExecutionError as e:
+        LOG.warning('Something went wrong when waiting for udev '
+                    'to settle. Error: %s', e)
